# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
from typing import Optional, Union

import pytest

import sciline as sl


def test_provider_returning_optional_allowed() -> None:
    def make_optional() -> int | None:
        return 3

    pl = sl.Pipeline([make_optional])
    assert pl.compute(int | None) == 3


def test_provider_returning_union_allowed() -> None:
    def make_union() -> int | float:
        return 3

    pl = sl.Pipeline([make_union])
    assert pl.compute(int | float) == 3


def test_parameter_type_union_or_optional_allowed() -> None:
    pipeline = sl.Pipeline()
    pipeline[int | float] = 3  # type: ignore[index]
    assert pipeline.compute(int | float) == 3
    pipeline[int | None] = 4  # type: ignore[index]
    assert pipeline.compute(int | None) == 4


def test_union_requirement_not_satisfied_by_any_of_its_arguments() -> None:
    def require_union(x: int | float) -> str:  # noqa: PYI041
        return f'{x}'

    pipeline = sl.Pipeline([require_union])
    pipeline[int] = 1
    with pytest.raises(sl.UnsatisfiedRequirement):
        pipeline.get(str)

    pipeline = sl.Pipeline([require_union])
    pipeline[float] = 1.2
    with pytest.raises(sl.UnsatisfiedRequirement):
        pipeline.get(str)


def test_optional_dependency_cannot_be_filled_by_non_optional_param() -> None:
    def use_optional(x: int | None) -> str:
        return f'{x or 123}'

    pipeline = sl.Pipeline([use_optional], params={int: 1})
    with pytest.raises(sl.UnsatisfiedRequirement):
        pipeline.get(str)


def test_optional_dependency_cannot_be_filled_by_non_optional_param_kwarg() -> None:
    def use_optional(*, x: int | None) -> str:
        return f'{x or 123}'

    pipeline = sl.Pipeline([use_optional], params={int: 1})
    with pytest.raises(sl.UnsatisfiedRequirement):
        pipeline.get(str)


def test_Union_argument_order_does_not_matter() -> None:
    def use_union(x: int | float) -> str:  # noqa: PYI041
        return f'{x}'

    pipeline = sl.Pipeline([use_union])
    pipeline[int | float] = 1  # type: ignore[index]
    assert pipeline.compute(str) == '1'
    pipeline = sl.Pipeline([use_union])
    pipeline[float | int] = 1  # type: ignore[index]
    assert pipeline.compute(str) == '1'
    # Note that the above works because the hashes are the same:
    assert hash(int | float) == hash(float | int)
    assert hash(Union[int, float]) == hash(Union[float, int])  # noqa: UP007


def test_optional_dependency_cannot_be_filled_transitively() -> None:
    def use_optional(x: int | None) -> str:
        return f'{x or 123}'

    def make_int(x: float) -> int:
        return int(x)

    pipeline = sl.Pipeline([use_optional, make_int], params={float: 2.2})
    with pytest.raises(sl.UnsatisfiedRequirement):
        pipeline.get(str)


def test_optional_dependency_can_be_set_to_None() -> None:
    def use_optional(x: int | None) -> str:
        return f'{x or 123}'

    pipeline = sl.Pipeline([use_optional])
    pipeline[Optional[int]] = None  # type: ignore[index] # noqa: UP007
    assert pipeline.compute(str) == '123'



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
"""Serialization of task graphs.

.. autosummary::
    :toctree: ../generated/functions
    :template: function-template.rst

    json_schema
    json_serialize_task_graph
"""

from ._json import json_schema, json_serialize_task_graph

__all__ = ['json_schema', 'json_serialize_task_graph']



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
from typing import TypeVar

from ._provider import Provider

T = TypeVar('T')


Key = type
Graph = dict[Key, Provider]


Json = dict[str, "Json"] | list["Json"] | str | int | float | bool | None



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
from __future__ import annotations

from collections.abc import Generator, Hashable, Sequence
from html import escape
from typing import Any, TypeVar

from ._utils import key_name
from .scheduler import DaskScheduler, NaiveScheduler, Scheduler
from .serialize import json_serialize_task_graph
from .typing import Graph, Json, Key

T = TypeVar("T")


def _list_items(items: Sequence[str]) -> str:
    return '\n'.join(
        (
            '<ul>',
            ('\n'.join(f'<li>{escape(it)}</li>' for it in items)),
            '</ul>',
        )
    )


def _list_max_n_then_hide(items: Sequence[str], n: int = 5, header: str = '') -> str:
    def wrap(s: str) -> str:
        return '\n'.join(
            (
                '<div class="task-graph-detail-list">'
                '<style> .task-graph-detail-list ul { margin-top: 0; } </style>',
                s,
                '</div>',
            )
        )

    return wrap(
        '\n'.join(
            (
                header,
                _list_items(items),
            )
        )
        if len(items) <= n
        else '\n'.join(
            (
                '<details>',
                '<style>',
                'details[open] .task-graph-summary ul { display: none; }',
                '</style>',
                '<summary class="task-graph-summary">',
                header,
                _list_items((*items[:n], '...')),
                '</summary>',
                _list_items(items),
                '</details>',
            )
        )
    )


Targets = Hashable | tuple[Hashable, ...]


class TaskGraph:
    """
    Holds a concrete task graph and keys to compute.

    Task graphs are typically created by :py:class:`sciline.Pipeline.build`. They allow
    for computing all or a subset of the results in the graph.
    """

    def __init__(
        self, *, graph: Graph, targets: Targets, scheduler: Scheduler | None = None
    ) -> None:
        self._graph = graph
        self._keys = targets
        if scheduler is None:
            try:
                scheduler = DaskScheduler()
            except ImportError:
                scheduler = NaiveScheduler()
        elif not isinstance(scheduler, Scheduler):
            raise ValueError(
                "Scheduler interface must be compatible with sciline.Scheduler"
            )
        self._scheduler = scheduler

    def compute(self, targets: Targets | None = None) -> Any:
        """
        Compute the result of the graph.

        Parameters
        ----------
        targets:
            Optional list of keys to compute. This can be used to override the keys
            stored in the graph instance. Note that the keys must be present in the
            graph as intermediate results, otherwise KeyError is raised.

        Returns
        -------
        If ``targets`` is a single type, returns the single result that was computed.
        If ``targets`` is a tuple of types, returns a dictionary with type as keys
        and the corresponding results as values.
        """
        if targets is None:
            targets = self._keys
        if isinstance(targets, tuple):
            results = self._scheduler.get(self._graph, list(targets))
            return dict(zip(targets, results, strict=True))
        else:
            return self._scheduler.get(self._graph, [targets])[0]

    def keys(self) -> Generator[Key, None, None]:
        """
        Iterate over all keys of the graph.

        Yields all keys, i.e., the types of values that can be computed or are
        provided as parameters.

        Returns
        -------
        :
            Iterable over keys.
        """
        yield from self._graph.keys()

    def visualize(self, **kwargs: Any) -> graphviz.Digraph:  # type: ignore[name-defined] # noqa: F821
        """
        Return a graphviz Digraph object representing the graph.

        Parameters
        ----------
        kwargs:
            Keyword arguments passed to :py:class:`graphviz.Digraph`.
        """
        from .visualize import to_graphviz

        return to_graphviz(self._graph, **kwargs)

    def serialize(self) -> dict[str, Json]:
        """Serialize the graph to JSON.

        See the user guide on
        `Serializing Providers <../../user-guide/serialization.rst>`_.

        Returns
        -------
        :
            A JSON object representing the graph.
        """
        return json_serialize_task_graph(self._graph)

    def _repr_html_(self) -> str:
        leafs = sorted(
            [
                escape(key_name(key))  # type: ignore[arg-type]
                for key in (
                    self._keys if isinstance(self._keys, tuple) else [self._keys]
                )
            ]
        )
        roots = sorted(
            {
                escape(key_name(key))
                for key, provider in self._graph.items()
                if provider.kind != 'function'
            }
        )
        scheduler = escape(str(self._scheduler))

        def head(word: str) -> str:
            return f'<h5>{word}</h5>'

        return '\n'.join(
            (
                '<style>.task-graph-repr h5 { display: inline; }</style>',
                '<div class="task-graph-repr">',
                head('Output keys: '),
                ','.join(leafs),
                '<br>',
                head('Scheduler: '),
                scheduler,
                '<br>',
                _list_max_n_then_hide(roots, header=head('Input keys:')),
                '</div>',
            )
        )



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
# type: ignore #noqa: PGH003

from copy import deepcopy
from typing import Any, NewType, TypeVar

import jsonschema
import pytest

import sciline as sl
from sciline.serialize import json_schema
from sciline.typing import Json

A = NewType('A', int)
B = NewType('B', int)
T = TypeVar('T', A, B)


class Int(sl.Scope[T, int], int): ...


class List(sl.Scope[T, list[int]], list[int]): ...


def make_int_b() -> Int[B]:
    return Int[B](2)


def zeros(n: Int[T]) -> List[T]:
    return List[T]([0] * n)


def to_string(b: List[B], a: List[A]) -> str:
    return f'a: {a}, b: {b}'


def as_float(x: int) -> float:
    return 0.5 * x


def make_graph_predictable(graph: dict[str, Json]) -> dict[str, Json]:
    """Sort nodes and edges and assign ids according to a fixed scheme.

    Nodes:
    - Sorted by ``out``.
    - Ids are counted up from 0 for the sorted nodes.

    Edges:
    - Sorted by ``source + target`` *after* remapping the node ids.
    - Ids are counted up from 100 for the sorted edges.
    """

    def node_sort_key(node: dict[str, Any]) -> str:
        if node['kind'] == 'data':
            return node['type'] + 'data'
        out_edge = next(e for e in graph['edges'] if e['source'] == node['id'])
        data_node = next(n for n in graph['nodes'] if n['id'] == out_edge['target'])
        return data_node['type'] + 'function'

    id_mapping = {}
    nodes: Any = sorted(deepcopy(graph['nodes']), key=node_sort_key)
    edges: Any = deepcopy(graph['edges'])
    for i, node in enumerate(nodes):
        new_id = str(i)
        id_mapping[node['id']] = new_id
        node['id'] = new_id

    for edge in edges:
        edge['source'] = id_mapping[edge['source']]
        edge['target'] = id_mapping[edge['target']]
    edges = sorted(edges, key=lambda e: e['source'] + e['target'])
    for i, edge in enumerate(edges, 100):
        new_id = str(i)
        id_mapping[edge['id']] = new_id
        edge['id'] = new_id

    for node in nodes:
        if node['kind'] == 'function':
            node['args'] = [id_mapping[arg] for arg in node['args']]
            node['kwargs'] = {k: id_mapping[v] for k, v in node['kwargs'].items()}

    return {**graph, 'nodes': nodes, 'edges': edges}


# Ids correspond to the result of assign_predictable_ids
expected_serialized_nodes = [
    {
        'id': '0',
        'label': 'str',
        'kind': 'data',
        'type': 'builtins.str',
    },
    {
        'id': '1',
        'label': 'to_string',
        'kind': 'function',
        'function': 'tests.serialize.json_test.to_string',
        'args': ['106', '104'],
        'kwargs': {},
    },
    {
        'id': '2',
        'label': 'Int[A]',
        'kind': 'data',
        'type': 'tests.serialize.json_test.Int[tests.serialize.json_test.A]',
    },
    {
        'id': '3',
        'label': 'Int[B]',
        'kind': 'data',
        'type': 'tests.serialize.json_test.Int[tests.serialize.json_test.B]',
    },
    {
        'id': '4',
        'label': 'make_int_b',
        'kind': 'function',
        'function': 'tests.serialize.json_test.make_int_b',
        'args': [],
        'kwargs': {},
    },
    {
        'id': '5',
        'label': 'List[A]',
        'kind': 'data',
        'type': 'tests.serialize.json_test.List[tests.serialize.json_test.A]',
    },
    {
        'id': '6',
        'label': 'zeros',
        'kind': 'function',
        'function': 'tests.serialize.json_test.zeros',
        'args': ['101'],
        'kwargs': {},
    },
    {
        'id': '7',
        'label': 'List[B]',
        'kind': 'data',
        'type': 'tests.serialize.json_test.List[tests.serialize.json_test.B]',
    },
    {
        'id': '8',
        'label': 'zeros',
        'kind': 'function',
        'function': 'tests.serialize.json_test.zeros',
        'args': ['102'],
        'kwargs': {},
    },
]
expected_serialized_edges = [
    {'id': '100', 'source': '1', 'target': '0'},
    {'id': '101', 'source': '2', 'target': '6'},
    {'id': '102', 'source': '3', 'target': '8'},
    {'id': '103', 'source': '4', 'target': '3'},
    {'id': '104', 'source': '5', 'target': '1'},
    {'id': '105', 'source': '6', 'target': '5'},
    {'id': '106', 'source': '7', 'target': '1'},
    {'id': '107', 'source': '8', 'target': '7'},
]
expected_serialized_graph = {
    'directed': True,
    'multigraph': False,
    'nodes': expected_serialized_nodes,
    'edges': expected_serialized_edges,
}


def test_serialize() -> None:
    pl = sl.Pipeline([make_int_b, zeros, to_string], params={Int[A]: 3})
    graph = pl.get(str)
    res = graph.serialize()
    res = make_graph_predictable(res)
    assert res == expected_serialized_graph


def fn_w_kwonlyargs(*, x: int) -> float:
    return 0.5 * x


# Ids correspond to the result of assign_predictable_ids
expected_serialized_kwonlyargs_nodes = [
    {
        'id': '0',
        'label': 'float',
        'kind': 'data',
        'type': 'builtins.float',
    },
    {
        'id': '1',
        'label': 'fn_w_kwonlyargs',
        'kind': 'function',
        'function': 'tests.serialize.json_test.fn_w_kwonlyargs',
        'args': [],
        'kwargs': {'x': '101'},
    },
    {
        'id': '2',
        'label': 'int',
        'kind': 'data',
        'type': 'builtins.int',
    },
]
expected_serialized_kwonlyargs_edges = [
    {'id': '100', 'source': '1', 'target': '0'},
    {'id': '101', 'source': '2', 'target': '1'},
]
expected_serialized_kwonlyargs_graph = {
    'directed': True,
    'multigraph': False,
    'nodes': expected_serialized_kwonlyargs_nodes,
    'edges': expected_serialized_kwonlyargs_edges,
}


def test_serialize_kwonlyargs() -> None:
    pl = sl.Pipeline([fn_w_kwonlyargs], params={int: 3})
    graph = pl.get(float)
    res = graph.serialize()
    res = make_graph_predictable(res)
    assert res == expected_serialized_kwonlyargs_graph


# Ids correspond to the result of assign_predictable_ids
expected_serialized_lambda_nodes = [
    {
        'id': '0',
        'label': 'float',
        'kind': 'data',
        'type': 'builtins.float',
    },
    {
        'id': '1',
        'label': '<lambda>',
        'kind': 'function',
        'function': 'tests.serialize.json_test.test_serialize_lambda.<locals>.<lambda>',
        'args': ['101'],
        'kwargs': {},
    },
    {
        'id': '2',
        'label': 'int',
        'kind': 'data',
        'type': 'builtins.int',
    },
]
expected_serialized_lambda_edges = [
    {'id': '100', 'source': '1', 'target': '0'},
    {'id': '101', 'source': '2', 'target': '1'},
]
expected_serialized_lambda_graph = {
    'directed': True,
    'multigraph': False,
    'nodes': expected_serialized_lambda_nodes,
    'edges': expected_serialized_lambda_edges,
}


def test_serialize_lambda() -> None:
    lam = lambda x: float(x)  # noqa: E731
    lam.__annotations__['x'] = int
    lam.__annotations__['return'] = float
    pl = sl.Pipeline((lam,), params={int: 4})
    graph = pl.get(float)
    res = graph.serialize()
    res = make_graph_predictable(res)
    assert res == expected_serialized_lambda_graph


def test_serialize_does_not_support_callable_objects() -> None:
    class C:
        def __call__(self, x: int) -> float:
            return float(x)

    pl = sl.Pipeline((C(),), params={int: 4})
    graph = pl.get(float)
    with pytest.raises(ValueError, match='Callable objects cannot be serialized'):
        graph.serialize()


def test_serialize_validate_schema() -> None:
    pl = sl.Pipeline([make_int_b, zeros, to_string], params={Int[A]: 3})
    graph = pl.get(str)
    res = graph.serialize()
    schema = json_schema()
    jsonschema.validate(res, schema)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
from typing import NewType

import pytest

import sciline as sl

A = NewType('A', int)
B = NewType('B', int)
C = NewType('C', int)
D = NewType('D', int)
X = NewType('X', int)


def a_to_b(a: A) -> B:
    return B(a + 1)


def b_to_c(b: B) -> C:
    return C(b + 2)


def c_to_d(c: C) -> D:
    return D(c + 4)


def test_map_returns_pipeline_that_can_compute_for_each_value() -> None:
    ab = sl.Pipeline((a_to_b,))
    mapped = ab.map({A: [A(10 * i) for i in range(3)]})
    with pytest.raises(sl.UnsatisfiedRequirement):
        # B is not in the graph any more, since it has been duplicated
        mapped.compute(B)
    from cyclebane.graph import IndexValues, NodeName

    for i in range(3):
        index = IndexValues(('dim_0',), (i,))
        assert mapped.compute(NodeName(A, index)) == A(10 * i)  # type: ignore[call-overload]
        assert mapped.compute(NodeName(B, index)) == B(10 * i + 1)  # type: ignore[call-overload]


def test_reduce_returns_pipeline_passing_mapped_branches_to_reducing_func() -> None:
    ab = sl.Pipeline((a_to_b,))
    mapped = ab.map({A: [A(10 * i) for i in range(3)]})
    # Define key to make mypy happy. This can actually be anything but currently
    # the type-hinting of Pipeline is too specific, disallowing, e.g., strings.
    Result = NewType('Result', int)
    assert mapped.reduce(func=min, name=Result).compute(Result) == Result(1)
    assert mapped.reduce(func=max, name=Result).compute(Result) == Result(21)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
import inspect
from collections import defaultdict
from collections.abc import Callable, Iterable
from typing import Any, TypeVar, get_args

from ._provider import Provider
from .typing import Key

T = TypeVar('T')
G = TypeVar('G')


def groupby(f: Callable[[T], G], a: Iterable[T]) -> defaultdict[G, list[T]]:
    g = defaultdict(list)
    for e in a:
        g[f(e)].append(e)
    return g


def full_qualname(obj: Any) -> str:
    try:
        obj_name = obj.__qualname__
    except AttributeError:
        obj_name = str(obj)

    module = inspect.getmodule(obj)
    if module is None:
        return str(obj_name)
    return f'{module.__name__}.{obj_name}'


def key_name(key: Key | TypeVar) -> str:
    args = get_args(key)
    if len(args):
        parameters = ', '.join(map(key_name, args))
        # getattr is a fallback for python < 3.10
        return f'{getattr(key, "__name__", "<Generic>")}[{parameters}]'
    if isinstance(key, TypeVar):
        return str(key)
    if hasattr(key, '__name__'):
        return key.__name__
    return str(key)


def key_full_qualname(key: Key | TypeVar) -> str:
    args = get_args(key)
    if len(args):
        origin = key.__origin__  # type: ignore[union-attr] # key is a TypeVar
        parameters = ', '.join(map(key_full_qualname, args))
        return f'{full_qualname(origin)}[{parameters}]'
    return full_qualname(key)


def provider_name(provider: Provider) -> str:
    if provider.kind == 'function':
        return provider.func.__name__
    if provider.kind == 'parameter':
        return f'parameter({key_name(type(provider.call({})))})'
    return str(provider)


def provider_full_qualname(provider: Provider) -> str:
    if provider.kind == 'function':
        return full_qualname(provider.func)
    if provider.kind == 'parameter':
        return f'parameter({key_full_qualname(type(provider.call({})))})'
    return str(provider)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
from __future__ import annotations

from collections.abc import Callable, Iterable
from itertools import chain
from types import UnionType
from typing import Any, TypeVar, get_args, get_type_hints, overload

from ._provider import Provider, ToProvider
from .data_graph import DataGraph, to_task_graph
from .display import pipeline_html_repr
from .handler import ErrorHandler, HandleAsComputeTimeException
from .scheduler import Scheduler
from .task_graph import TaskGraph
from .typing import Key

T = TypeVar('T')
KeyType = TypeVar('KeyType', bound=Key)


def _is_multiple_keys(keys: type | Iterable[type] | UnionType) -> bool:
    # Cannot simply use isinstance(keys, Iterable) because that is True for
    # generic aliases of iterable types, e.g.,
    #
    # class Str(sl.Scope[Param, str], str): ...
    # keys = Str[int]
    #
    # And isinstance(keys, type) does not work on its own because
    # it is False for the above type.
    if isinstance(keys, str):
        return False
    return (
        not isinstance(keys, type) and not get_args(keys) and isinstance(keys, Iterable)
    )


class Pipeline(DataGraph):
    """A container for providers that can be assembled into a task graph."""

    def __init__(
        self,
        providers: Iterable[ToProvider | Provider] | None = None,
        *,
        params: dict[type[Any], Any] | None = None,
    ):
        """
        Setup a Pipeline from a list providers

        Parameters
        ----------
        providers:
            List of callable providers. Each provides its return value.
            Their arguments and return value must be annotated with type hints.
        params:
            Dictionary of concrete values to provide for types.
        """
        super().__init__(providers)
        for tp, param in (params or {}).items():
            self[tp] = param

    @overload
    def compute(self, tp: type[T], **kwargs: Any) -> T: ...

    @overload
    def compute(self, tp: Iterable[type[T]], **kwargs: Any) -> dict[type[T], T]: ...

    @overload
    def compute(self, tp: UnionType, **kwargs: Any) -> Any: ...

    def compute(self, tp: type | Iterable[type] | UnionType, **kwargs: Any) -> Any:
        """
        Compute result for the given keys.

        Equivalent to ``self.get(tp).compute()``.

        Parameters
        ----------
        tp:
            Type to compute the result for.
            Can be a single type or an iterable of types.
        kwargs:
            Keyword arguments passed to the ``.get()`` method.
        """
        return self.get(tp, **kwargs).compute()

    def visualize(self, tp: type | Iterable[type], **kwargs: Any) -> graphviz.Digraph:  # type: ignore[name-defined] # noqa: F821
        """
        Return a graphviz Digraph object representing the graph for the given keys.

        Equivalent to ``self.get(tp).visualize()``.

        Parameters
        ----------
        tp:
            Type to visualize the graph for.
            Can be a single type or an iterable of types.
        kwargs:
            Keyword arguments passed to :py:class:`graphviz.Digraph`.
        """
        return self.get(tp, handler=HandleAsComputeTimeException()).visualize(**kwargs)

    def get(
        self,
        keys: type | Iterable[type] | UnionType,
        *,
        scheduler: Scheduler | None = None,
        handler: ErrorHandler | None = None,
    ) -> TaskGraph:
        """
        Return a TaskGraph for the given keys.

        Parameters
        ----------
        keys:
            Type to compute the result for.
            Can be a single type or an iterable of types.
        scheduler:
            Optional scheduler to use for computing the result. If not given, a
            :py:class:`NaiveScheduler` is used if `dask` is not installed,
            otherwise dask's threaded scheduler is used.
        handler:
            Handler for unsatisfied requirements. If not provided,
            :py:class:`HandleAsBuildTimeException` is used, which raises an exception.
            During development and debugging it can be helpful to use a handler that
            raises an exception only when the graph is computed. This can be achieved
            by passing :py:class:`HandleAsComputeTimeException` as the handler.
        """
        if multi := _is_multiple_keys(keys):
            targets = tuple(keys)  # type: ignore[arg-type]
        else:
            targets = (keys,)  # type: ignore[assignment]
        graph = to_task_graph(self, targets=targets, handler=handler)
        return TaskGraph(
            graph=graph,
            targets=targets if multi else keys,  # type: ignore[arg-type]
            scheduler=scheduler,
        )

    @overload
    def bind_and_call(self, fns: Callable[..., T], /) -> T: ...

    @overload
    def bind_and_call(
        self, fns: Iterable[Callable[..., Any]], /
    ) -> tuple[Any, ...]: ...

    def bind_and_call(
        self, fns: Callable[..., Any] | Iterable[Callable[..., Any]], /
    ) -> Any:
        """
        Call the given functions with arguments provided by the pipeline.

        Parameters
        ----------
        fns:
            Functions to call.
            The pipeline will provide all arguments based on the function's type hints.

            If this is a single callable, it is called directly.
            Otherwise, ``bind_and_call`` will iterate over it and call all functions.
            If will in either case call :meth:`Pipeline.compute` only once.

        Returns
        -------
        :
            The return values of the functions in the same order as the functions.
            If only one function is passed, its return value
            is *not* wrapped in a tuple.
        """
        return_tuple = True
        if callable(fns):
            fns = (fns,)
            return_tuple = False

        arg_types_per_function = {
            fn: {
                name: ty for name, ty in get_type_hints(fn).items() if name != 'return'
            }
            for fn in fns
        }
        all_arg_types = tuple(
            set(chain(*(a.values() for a in arg_types_per_function.values())))
        )
        values_per_type = self.compute(all_arg_types)
        results = tuple(
            fn(**{name: values_per_type[ty] for name, ty in arg_types.items()})
            for fn, arg_types in arg_types_per_function.items()
        )
        if not return_tuple:
            return results[0]
        return results

    def _repr_html_(self) -> str:
        nodes = ((key, data) for key, data in self._graph.nodes.items())
        return pipeline_html_repr(nodes)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
from typing import NewType

import pytest

import sciline as sl

A = NewType('A', int)
B = NewType('B', int)
C = NewType('C', int)
D = NewType('D', int)
X = NewType('X', int)


def a_to_b(a: A) -> B:
    return B(a + 1)


def ab_to_c(a: A, b: B) -> C:
    return C(a + b)


def b_to_c(b: B) -> C:
    return C(b + 2)


def c_to_d(c: C) -> D:
    return D(c + 4)


def test_setitem_can_compose_pipelines() -> None:
    ab = sl.Pipeline((a_to_b,))
    ab[A] = 0
    bc = sl.Pipeline((b_to_c,))
    bc[B] = ab
    assert bc.compute(C) == C(3)


def test_setitem_raises_if_value_pipeline_has_no_unique_output() -> None:
    abx = sl.Pipeline((a_to_b,))
    abx[X] = 666
    bc = sl.Pipeline((b_to_c,))
    with pytest.raises(ValueError, match='Graph must have exactly one sink node'):
        bc[B] = abx


def test_setitem_can_add_value_pipeline_at_new_node_given_by_key() -> None:
    ab = sl.Pipeline((a_to_b,))
    ab[A] = 0
    empty = sl.Pipeline(())
    empty[B] = ab
    assert empty.compute(B) == B(1)


def test_setitem_replaces_existing_node_value_with_value_pipeline() -> None:
    ab = sl.Pipeline((a_to_b,))
    ab[A] = 0
    bc = sl.Pipeline((b_to_c,))
    bc[B] = 111
    bc[B] = ab
    assert bc.compute(C) == C(3)


def test_setitem_replaces_existing_branch_with_value() -> None:
    ab = sl.Pipeline((a_to_b,))
    ab[A] = 0
    bc = sl.Pipeline((b_to_c,))
    bc[B] = ab
    bc[B] = 111
    assert bc.compute(C) == C(113)
    # __setitem__ prunes the entire branch, instead of just cutting the edges
    with pytest.raises(sl.UnsatisfiedRequirement):
        bc.compute(A)


def test_setitem_with_conflicting_nodes_in_value_pipeline_raises_on_data_mismatch() -> (
    None
):
    ab = sl.Pipeline((a_to_b,))
    ab[A] = 100
    abc = sl.Pipeline((ab_to_c,))
    abc[A] = 666
    with pytest.raises(ValueError, match="Node data differs"):
        abc[B] = ab


def test_setitem_with_conflicting_nodes_in_value_pipeline_accepts_on_data_match() -> (
    None
):
    ab = sl.Pipeline((a_to_b,))
    ab[A] = 100
    abc = sl.Pipeline((ab_to_c,))
    abc[A] = 100
    abc[B] = ab
    assert abc.compute(C) == C(201)


def test_setitem_with_conflicting_nodes_in_value_pipeline_raises_on_unique_data() -> (
    None
):
    ab = sl.Pipeline((a_to_b,))
    ab[A] = 100
    abc = sl.Pipeline((ab_to_c,))
    # Missing data is just as bad as conflicting data. At first glance, it might seem
    # like this should be allowed, but it would be a source of bugs and confusion.
    # In particular since it would allow for adding providers or parents to an
    # indirect dependency of the key in setitem, which would be very confusing.
    with pytest.raises(ValueError, match="Node data differs"):
        abc[B] = ab


def test_setitem_with_conflicting_node_inputs_in_value_pipeline_raises() -> None:
    def x_to_b(x: X) -> B:
        return B(x + 1)

    def bc_to_d(b: B, c: C) -> D:
        return D(b + c)

    xbc = sl.Pipeline((x_to_b, b_to_c))
    xbc[X] = 666
    abcd = sl.Pipeline((a_to_b, bc_to_d))
    abcd[A] = 100
    with pytest.raises(ValueError, match="Node inputs differ"):
        # If this worked naively by composing the NetworkX graph, it would look as
        # below, giving B 2 inputs instead of 1, even though provider has 1 argument,
        # corrupting the graph:
        # A  X
        # \ /
        #  B
        #  |\
        #  C |
        #  |/
        #  D
        abcd[C] = xbc



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
from __future__ import annotations

import importlib.resources
import json
from typing import Any

from .._provider import Provider
from .._utils import key_full_qualname, key_name, provider_full_qualname, provider_name
from ..typing import Graph, Json, Key


def json_schema() -> Json:
    """Return the JSON schema for serialized task graphs.

    See, e.g., `jsonschema <https://python-jsonschema.readthedocs.io/en/stable/>`_
    for a tool that can validate the schema.

    Returns
    -------
    :
        The graph JSON schema as a dict.
    """
    with (
        importlib.resources.files('sciline.serialize')
        .joinpath('graph_json_schema.json')
        .open()
    ) as f:
        return json.load(f)  # type: ignore[no-any-return]


def json_serialize_task_graph(graph: Graph) -> dict[str, Json]:
    """Serialize a graph to JSON.

    See the user guide on
    `Serializing Providers <../../user-guide/serialization.rst>`_.

    Also available as :meth:`sciline.TaskGraph.serialize`.

    Returns
    -------
    :
        A JSON object representing the graph.
    """
    id_gen = _IdGenerator()
    nodes = []
    edges = []
    for key, provider in graph.items():
        n, e = _serialize_provider(key, provider, id_gen)
        nodes.extend(n)
        edges.extend(e)
    return {
        'directed': True,
        'multigraph': False,
        'nodes': nodes,  # type: ignore[dict-item]
        'edges': edges,  # type: ignore[dict-item]
    }


# returns tuple[list[node], list[edge]]
# where the nodes are either a single data node or a data node and a function node
def _serialize_provider(
    key: Key, provider: Provider, id_gen: _IdGenerator
) -> tuple[list[dict[str, Json]], list[dict[str, Json]]]:
    if provider.kind == 'function':
        return _serialize_function(key, provider, id_gen)
    if provider.kind == 'parameter':
        return _serialize_param(key, id_gen)
    raise ValueError(
        f'Cannot serialize a task graph that contains {provider.kind} nodes.'
    )


def _serialize_param(
    key: Key, id_gen: _IdGenerator
) -> tuple[list[dict[str, Json]], list[dict[str, Json]]]:
    node = {
        'id': id_gen.data_node_id(key),
        'kind': 'data',
        'label': key_name(key),
        'type': key_full_qualname(key),
    }
    return [node], []  # type: ignore[list-item]


def _serialize_function(
    key: Key, provider: Provider, id_gen: _IdGenerator
) -> tuple[list[dict[str, Json]], list[dict[str, Json]]]:
    edges = []
    args = []
    kwargs = {}
    for i, arg in enumerate(provider.arg_spec.args):
        edge = _serialize_edge_data_to_fn(arg, key, i, id_gen)
        edges.append(edge)
        args.append(edge['id'])
    for name, kwarg in provider.arg_spec.kwargs:
        edge = _serialize_edge_data_to_fn(kwarg, key, name, id_gen)
        edges.append(edge)
        kwargs[name] = edge['id']
    edges.append(_serialize_edge_fn_to_data(key, key, id_gen))

    fn_node = {
        'id': id_gen.function_node_id(key),
        'kind': 'function',
        'label': _provider_name(provider),
        'function': provider_full_qualname(provider),
        'args': args,
        'kwargs': kwargs,
    }
    [data_node], _ = _serialize_param(key, id_gen)

    return [fn_node, data_node], edges  # type: ignore[list-item, return-value]


def _serialize_edge_data_to_fn(
    source: Key, target: Key, arg: int | str, id_gen: _IdGenerator
) -> dict[str, str]:
    return {
        'id': id_gen.edge_id(source, target, arg),
        'source': id_gen.data_node_id(source),
        'target': id_gen.function_node_id(target),
    }


def _serialize_edge_fn_to_data(
    source: Key, target: Key, id_gen: _IdGenerator
) -> dict[str, str]:
    return {
        'id': id_gen.edge_id(source, target, 0),
        'source': id_gen.function_node_id(source),
        'target': id_gen.data_node_id(target),
    }


def _provider_name(provider: Provider) -> str:
    try:
        return provider_name(provider)
    except AttributeError:
        # E.g. with callable objects:
        # AttributeError: '<class>' object has no attribute '__name__'
        raise ValueError(
            f"Unsupported provider for serializing graph: '{provider}' "
            'Callable objects cannot be serialized.'
        ) from None


class _IdGenerator:
    def __init__(self) -> None:
        self._assigned: dict[Any, str] = {}
        self._next = 0

    def data_node_id(self, key: Key) -> str:
        # Keys must be unique and are required to be hashable to construct TaskGraph.
        return self._get_or_insert(key)

    def function_node_id(self, key: Key) -> str:
        # Use the key instead of the provider to avoid problems with
        # unhashable providers.
        # Use tuple with 'fn' to disambiguate from data nodes.
        return self._get_or_insert(('fn', key))

    def edge_id(self, source: Key, target: Key, arg: int | str) -> str:
        # Uses the arg number or kwarg name
        # to disambiguate arguments with the same type.
        return self._get_or_insert((source, target, arg))

    def _get_or_insert(self, hashable: Any) -> str:
        try:
            return self._assigned[hashable]
        except KeyError:
            id_ = str(self._next)
            self._assigned[hashable] = id_
            self._next += 1
            return id_



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024 Scipp contributors (https://github.com/scipp)
from __future__ import annotations

import itertools
from collections.abc import Callable, Generator, Iterable
from types import NoneType
from typing import Any, TypeVar, get_args

import cyclebane as cb
import networkx as nx

from ._provider import ArgSpec, Provider, ToProvider, _bind_free_typevars
from .handler import ErrorHandler, HandleAsBuildTimeException
from .typing import Graph, Key


def _as_graph(key: Key, value: Any) -> cb.Graph:
    """Create a cyclebane.Graph with a single value."""
    graph = nx.DiGraph()
    graph.add_node(key, value=value)
    return cb.Graph(graph)


def _find_all_typevars(t: type | TypeVar) -> set[TypeVar]:
    """Returns the set of all TypeVars in a type expression."""
    if isinstance(t, TypeVar):
        return {t}
    return set(itertools.chain(*map(_find_all_typevars, get_args(t))))


def _get_typevar_constraints(t: TypeVar) -> set[type]:
    """Returns the set of constraints of a TypeVar."""
    return set(t.__constraints__)


def _mapping_to_constrained(
    type_vars: set[TypeVar],
) -> Generator[dict[TypeVar, type], None, None]:
    constraints = [_get_typevar_constraints(t) for t in type_vars]
    if any(len(c) == 0 for c in constraints):
        raise ValueError('Typevars must have constraints')
    for combination in itertools.product(*constraints):
        yield dict(zip(type_vars, combination, strict=True))


T = TypeVar('T', bound='DataGraph')


class DataGraph:
    def __init__(self, providers: None | Iterable[ToProvider | Provider]) -> None:
        self._cbgraph = cb.Graph(nx.DiGraph())
        for provider in providers or []:
            self.insert(provider)

    @classmethod
    def _from_cyclebane(cls: type[T], graph: cb.Graph) -> T:
        out = cls([])
        out._cbgraph = graph
        return out

    def copy(self: T) -> T:
        return self._from_cyclebane(self._cbgraph.copy())

    def __copy__(self: T) -> T:
        return self.copy()

    @property
    def _graph(self) -> nx.DiGraph:
        return self._cbgraph.graph

    def _get_clean_node(self, key: Key) -> Any:
        """Return node ready for setting value or provider."""
        if key is NoneType:
            raise ValueError('Key must not be None')
        if key in self._graph:
            self._graph.remove_edges_from(list(self._graph.in_edges(key)))
            self._graph.nodes[key].pop('value', None)
            self._graph.nodes[key].pop('provider', None)
            self._graph.nodes[key].pop('reduce', None)
        else:
            self._graph.add_node(key)
        return self._graph.nodes[key]

    def insert(self, provider: ToProvider | Provider, /) -> None:
        """
        Insert a callable into the graph that provides its return value.

        Parameters
        ----------
        provider:
            Either a callable that provides its return value. Its arguments
            and return value must be annotated with type hints.
            Or a ``Provider`` object that has been constructed from such a callable.
        """
        if not isinstance(provider, Provider):
            provider = Provider.from_function(provider)
        return_type = provider.deduce_key()
        if typevars := _find_all_typevars(return_type):
            for bound in _mapping_to_constrained(typevars):
                self.insert(provider.bind_type_vars(bound))
            return
        # Trigger UnboundTypeVar error if any input typevars are not bound
        provider = provider.bind_type_vars({})
        self._get_clean_node(return_type)['provider'] = provider
        for dep in provider.arg_spec.keys():
            self._graph.add_edge(dep, return_type, key=dep)

    def __setitem__(self, key: Key, value: DataGraph | Any) -> None:
        """
        Provide a concrete value for a type.

        Parameters
        ----------
        key:
            Type to provide a value for.
        param:
            Concrete value to provide.
        """
        # This is a questionable approach: Using MyGeneric[T] as a key will actually
        # not pass mypy [valid-type] checks. What we do on our side is ok, but the
        # calling code is not.
        if typevars := _find_all_typevars(key):
            for bound in _mapping_to_constrained(typevars):
                self[_bind_free_typevars(key, bound)] = value
            return

        # TODO If key is generic, should we support multi-sink case and update all?
        # Would imply that we need the same for __getitem__.
        self._cbgraph[key] = (
            value._cbgraph if isinstance(value, DataGraph) else _as_graph(key, value)
        )

    def __getitem__(self: T, key: Key) -> T:
        return self._from_cyclebane(self._cbgraph[key])

    def map(self: T, node_values: dict[Key, Any]) -> T:
        return self._from_cyclebane(self._cbgraph.map(node_values))

    def reduce(self: T, *, func: Callable[..., Any], **kwargs: Any) -> T:
        # Note that the type hints of `func` are not checked here. As we are explicit
        # about the modification, this is in line with __setitem__ which does not
        # perform such checks and allows for using generic reduction functions.
        return self._from_cyclebane(
            self._cbgraph.reduce(attrs={'reduce': func}, **kwargs)
        )

    def to_networkx(self) -> nx.DiGraph:
        return self._cbgraph.to_networkx()

    def visualize_data_graph(self, **kwargs: Any) -> graphviz.Digraph:  # type: ignore[name-defined] # noqa: F821
        import graphviz

        dot = graphviz.Digraph(strict=True, **kwargs)
        for node in self._graph.nodes:
            dot.node(str(node), label=str(node), shape='box')
            attrs = self._graph.nodes[node]
            attrs = '\n'.join(f'{k}={v}' for k, v in attrs.items())
            dot.node(str(node), label=f'{node}\n{attrs}', shape='box')
        for edge in self._graph.edges:
            key = self._graph.edges[edge].get('key')
            label = str(key) if key is not None else ''
            dot.edge(str(edge[0]), str(edge[1]), label=label)
        return dot


_no_value = object()


def to_task_graph(
    data_graph: DataGraph, targets: tuple[Key, ...], handler: ErrorHandler | None = None
) -> Graph:
    graph = data_graph.to_networkx()
    handler = handler or HandleAsBuildTimeException()
    ancestors = list(targets)
    for target in targets:
        if target not in graph:
            handler.handle_unsatisfied_requirement(target)
        ancestors.extend(nx.ancestors(graph, target))
    graph = graph.subgraph(set(ancestors))
    out = {}

    for key in graph.nodes:
        node = graph.nodes[key]
        input_nodes = list(graph.predecessors(key))
        input_edges = list(graph.in_edges(key, data=True))
        orig_keys = [edge[2].get('key', None) for edge in input_edges]
        if (value := node.get('value', _no_value)) is not _no_value:
            out[key] = Provider.parameter(value)
        elif (provider := node.get('provider')) is not None:
            new_key = dict(zip(orig_keys, input_nodes, strict=True))
            spec = provider.arg_spec.map_keys(new_key.get)
            if len(spec) != len(input_nodes):
                # This should be caught by __setitem__, but we check here to be safe.
                raise ValueError("Corrupted graph")
            # TODO also kwargs
            out[key] = Provider(func=provider.func, arg_spec=spec, kind='function')
        elif (func := node.get('reduce')) is not None:
            spec = ArgSpec.from_args(*input_nodes)
            out[key] = Provider(func=func, arg_spec=spec, kind='function')
        else:
            out[key] = handler.handle_unsatisfied_requirement(key)
    return out



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
import sys
from typing import NewType, TypeVar

import pytest

import sciline
from sciline import _utils
from sciline._provider import Provider


def module_foo(x: list[str]) -> str:
    return x[0]


def test_key_name_builtin() -> None:
    assert _utils.key_name(int) == 'int'
    assert _utils.key_name(object) == 'object'


def test_key_name_new_type() -> None:
    MyType = NewType('MyType', str)
    assert _utils.key_name(MyType) == 'MyType'


@pytest.mark.skipif(sys.version_info < (3, 12), reason="requires python3.12 or higher")
def test_key_name_type_alias_type() -> None:
    # Use exec to avoid a syntax error in older python
    code = """
type MyType = float
assert _utils.key_name(MyType) == 'MyType'
    """
    exec(code)  # noqa: S102


@pytest.mark.skipif(sys.version_info < (3, 12), reason="requires python3.12 or higher")
def test_key_name_generic_type_alias_type() -> None:
    # Use exec to avoid a syntax error in older python
    code = """
type MyType[T] = dict[str, T]
assert _utils.key_name(MyType[int]) == 'MyType[int]'
    """
    exec(code)  # noqa: S102


def test_key_name_type_var() -> None:
    MyType = TypeVar('MyType')
    assert _utils.key_name(MyType) == '~MyType'  # type: ignore[arg-type]


def test_key_name_builtin_generic() -> None:
    MyType = NewType('MyType', str)
    assert _utils.key_name(list) == 'list'
    assert _utils.key_name(list[float]) == 'list[float]'
    assert _utils.key_name(list[MyType]) == 'list[MyType]'
    assert _utils.key_name(dict[str, MyType]) == 'dict[str, MyType]'


def test_key_name_custom_generic() -> None:
    MyType = NewType('MyType', float)
    Var = TypeVar('Var')

    class G(sciline.Scope[Var, str], str): ...

    assert _utils.key_name(G) == 'G'
    assert _utils.key_name(G[int]) == 'G[int]'
    assert _utils.key_name(G[MyType]) == 'G[MyType]'


def test_key_name_custom_generic_two_params() -> None:
    MyType = NewType('MyType', float)
    Var1 = TypeVar('Var1')
    Var2 = TypeVar('Var2')

    class G(sciline.ScopeTwoParams[Var1, Var2, str], str): ...

    assert _utils.key_name(G) == 'G'
    assert _utils.key_name(G[int, tuple[float]]) == 'G[int, tuple[float]]'
    assert _utils.key_name(G[list[MyType], MyType]) == 'G[list[MyType], MyType]'


def test_key_full_qualname_builtin() -> None:
    assert _utils.key_full_qualname(int) == 'builtins.int'
    assert _utils.key_full_qualname(object) == 'builtins.object'


def test_key_full_qualname_new_type() -> None:
    # The __qualname__ of NewTypes is the same as __name__, the result is therefore
    # missing test_key_full_qualname_new_type.<locals>
    MyType = NewType('MyType', str)
    assert _utils.key_full_qualname(MyType) == 'tests.utils_test.MyType'


@pytest.mark.skipif(sys.version_info < (3, 12), reason="requires python3.12 or higher")
def test_key_full_qualname_type_alias_type() -> None:
    # Use exec to avoid a syntax error in older python
    code = """
type MyType = float
assert _utils.key_full_qualname(MyType) == 'tests.utils_test.MyType'
    """
    exec(code)  # noqa: S102


@pytest.mark.skipif(sys.version_info < (3, 12), reason="requires python3.12 or higher")
def test_key_full_qualname_generic_type_alias_type() -> None:
    # Use exec to avoid a syntax error in older python
    code = """
type MyType[T] = dict[str, T]
assert _utils.key_full_qualname(MyType[int]) == 'tests.utils_test.MyType[builtins.int]'
    """
    exec(code)  # noqa: S102


def test_key_full_qualname_type_var() -> None:
    # TypeVar has no __qualname__, the result is therefore missing
    # test_key_full_qualname_type_var.<locals>
    MyType = TypeVar('MyType')
    res = _utils.key_full_qualname(MyType)  # type: ignore[arg-type]
    assert res == 'tests.utils_test.~MyType'


@pytest.mark.skipif(sys.version_info < (3, 10), reason="requires python3.10 or higher")
def test_key_full_qualname_builtin_generic() -> None:
    MyType = NewType('MyType', str)
    assert _utils.key_full_qualname(list) == 'builtins.list'
    assert _utils.key_full_qualname(list[float]) == 'builtins.list[builtins.float]'
    assert (
        _utils.key_full_qualname(list[MyType])
        == 'builtins.list[tests.utils_test.MyType]'
    )
    assert (
        _utils.key_full_qualname(dict[str, MyType])
        == 'builtins.dict[builtins.str, tests.utils_test.MyType]'
    )


def test_key_full_qualname_custom_generic() -> None:
    MyType = NewType('MyType', float)
    Var = TypeVar('Var')

    class G(sciline.Scope[Var, str], str): ...

    assert (
        _utils.key_full_qualname(G)
        == 'tests.utils_test.test_key_full_qualname_custom_generic.<locals>.G'
    )
    assert (
        _utils.key_full_qualname(G[int])
        == 'tests.utils_test.test_key_full_qualname_custom_generic.<locals>.G[builtins.int]'  # noqa: E501
    )
    assert (
        _utils.key_full_qualname(G[MyType])
        == 'tests.utils_test.test_key_full_qualname_custom_generic.'
        '<locals>.G[tests.utils_test.MyType]'
    )


def test_key_full_qualname_custom_generic_two_params() -> None:
    MyType = NewType('MyType', float)
    Var1 = TypeVar('Var1')
    Var2 = TypeVar('Var2')

    class G(sciline.ScopeTwoParams[Var1, Var2, str], str): ...

    assert (
        _utils.key_full_qualname(G)
        == 'tests.utils_test.test_key_full_qualname_custom_generic_two_params.<locals>.G'  # noqa: E501
    )
    assert (
        _utils.key_full_qualname(G[int, tuple[float]])
        == 'tests.utils_test.test_key_full_qualname_custom_generic_two_params.'
        '<locals>.G[builtins.int, builtins.tuple[builtins.float]]'
    )
    assert (
        _utils.key_full_qualname(G[list[MyType], MyType])
        == 'tests.utils_test.test_key_full_qualname_custom_generic_two_params.<locals>.'
        'G[builtins.list[tests.utils_test.MyType], tests.utils_test.MyType]'
    )


def test_provider_name_function() -> None:
    def foo(i: int) -> float:
        return float(i)

    assert _utils.provider_name(Provider.from_function(foo)) == 'foo'
    assert _utils.provider_name(Provider.from_function(module_foo)) == 'module_foo'


def test_provider_name_parameter() -> None:
    MyType = NewType('MyType', float)
    assert _utils.provider_name(Provider.parameter(4.1)) == 'parameter(float)'
    assert _utils.provider_name(Provider.parameter(MyType(3.2))) == 'parameter(float)'


def test_provider_full_qualname_function() -> None:
    def foo(i: int) -> float:
        return float(i)

    assert (
        _utils.provider_full_qualname(Provider.from_function(foo))
        == 'tests.utils_test.test_provider_full_qualname_function.<locals>.foo'
    )
    assert (
        _utils.provider_full_qualname(Provider.from_function(module_foo))
        == 'tests.utils_test.module_foo'
    )


def test_provider_full_qualname_parameter() -> None:
    MyType = NewType('MyType', float)
    assert _utils.provider_name(Provider.parameter(4.1)) == 'parameter(float)'
    assert _utils.provider_name(Provider.parameter(MyType(3.2))) == 'parameter(float)'



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
from collections.abc import Iterable
from html import escape
from typing import Any

from ._utils import key_name
from .typing import Key


def _details(summary: str, body: str) -> str:
    return f'''
    <details>
      <summary>{summary}</summary>
      {body}
    </details>
    '''


def _provider_name(p: tuple[Key, dict[str, Any]]) -> str:
    return escape(key_name(p[0]))


def _provider_source(data: dict[str, Any]) -> str:
    if (v := data.get('provider', None)) is None:
        return ''
    if v.kind == 'function':
        return _details(
            escape(v.location.name),
            escape(f'{v.location.module}.{v.location.name}'),
        )
    return ''


def _provider_value(data: dict[str, Any]) -> str:
    if (value := data.get('value', None)) is None:
        return ''
    html = escape(str(value)).strip()
    return _details(f'{html[:30]}...', html) if len(html) > 30 else html


def pipeline_html_repr(nodes: Iterable[tuple[Key, dict[str, Any]]]) -> str:
    provider_rows = '\n'.join(
        f'''
        <tr>
          <td scope="row">{_provider_name(item)}</td>
          <td scope="row">{_provider_value(item[1])}</td>
          <td scope="row">{_provider_source(item[1])}</th>
        </tr>'''
        for item in sorted(nodes, key=_provider_name)
    )
    return f'''
    <div class="pipeline-html-repr">
      <table>
        <thead>
          <tr>
            <th scope="col">Name</th>
            <th scope="col">Value</th>
            <th scope="col">Source</th>
          </tr>
        </thead>
        <tbody>
          {provider_rows}
        </tbody>
      </table>
    </div>
    '''.strip()
